%% Fixpoint typer for Sloth
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% Originally written by: (No code is left, only ideas)
%%   * Julio Mariño Carballo
%%   * Jose María Rey
%%   * Miguel Ángel Polo
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(typer, [type_module/1, infer_type_expr/2] ).

:- use_module(library(lists),       [append/3, difference/3]).
:- use_module(library(terms_check), [instance/2]).

:- use_module(symtab,   [get_rules/2, set_object_property/3, get_object_property/3,
                         get_module_functions/2] ).
:- use_module(curryast, [split_rule/3, get_function/2, get_params/2,
                         map_tree_list/3] ).
:- use_module(cgraph,   [build_cgraph/2, cgraph_get_visited/2] ).

:- use_module(types_unify, [unify_types/2]).

%% Operators includes.
:- include(type_ops).
:- include(ast_ops).

%% The typer operates directly with the ast representation given by the parser.
%% 
%% 1.- Get call graph for the modules.
%% 2.- Type in reverse call order. If we have fixpoint typing, replace
%%     the fixpoint functions by variables and type that.

%% Predicates starting with infer_* mean that they do actual type
%% inference. The ones starting with type_* do type checking also.

%% Expression typer.
%% infer_type(Expr, Type) succeeds if Expr has type Type.

%% Small hack, constants already carry their type.
infer_type_expr(const(Type, _Value), Type) :- 
	!.

%% Variables have a to be known type.
infer_type_expr(freevar(X), X) :- !.
infer_type_expr(param(X), X) :- !.

%% Identifiers should already have a type.
infer_type_expr(qid(M, N), Type) :-
	!,
	(
	    get_object_property(qid(M,N), type, Type) ->
	    true
	;
	    throw(error(typer, missing_type(qid(M, N))))
	).	
%% For conditional exprs. It's a little bit ugly but the best way
%% we've got to not touch so much our existing infrastructure.
infer_type_expr(untyped_cond(List), Type) :-
	infer_type_expr(List, CType),
	CType = tid('Prelude', [])<@[tid('Prelude', tuple)<@[_CondT, BodyT]],
	Type = BodyT.

%% For function applications.
infer_type_expr(F@Args, Type) :-
	infer_type_expr(F, FType),
	infer_check_fapp(FType, Args, Type).

%% infer_check_fapp checks a function application.
infer_check_fapp(FType, [], FType) :- !.
infer_check_fapp(FType, [A|AS], RType) :-
	(
	    FType = (T1 --> T2),
	    infer_type_expr(A, F1Type),
	    unify_types(T1, F1Type),
	    !,
	    infer_check_fapp(T2, AS, RType)
	;
	    %% Bad, we are applying too much arguments.
	    throw(error(typer, too_much_args(A)))
	).

infer_type_list([], []).
infer_type_list([E|ER], [T|TR]) :-
	infer_type_expr(E, T),
	!,
	infer_type_list(ER, TR).

%% Rules inference.
infer_type_rule(Rule, Type) :-
	split_rule(Rule, Head, Body),
	get_params(Head, LParams),
	infer_type_list(LParams, HeadTypes),
	infer_type_expr(Body, BodyType),
	append(HeadTypes, [BodyType], TotalType),
	list2type(TotalType, Type),
	get_function(Head, Func),
	(
	    \+ground(Func) -> %% This should handle the fixpoint case.
	    infer_type_expr(Func, FT),
	    unify_types(FT, Type)
	;
	    true
	).

infer_type_rules([], []).
infer_type_rules([R|RS], [T|TS]) :-
	infer_type_rule(R, T),
	!,
	infer_type_rules(RS, TS).

%% This is repeated in some places, FIXME, refactorize.
list2type([A], A).
list2type([A|LR], A --> TR) :-
	list2type(LR, TR).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Module typer.
%% Check all types for a given module.

type_module(Module) :-
	get_module_functions(Module, Funcs),
	type_flist(Funcs).

type_flist([]).
type_flist([F|FR]) :-
	build_cgraph(F, CG),
	!,
	type_cgraph(CG),
	cgraph_get_visited(CG, TypedFuncs),
	difference(FR, TypedFuncs, RestFunc),
	type_flist(RestFunc).

%% Cgraph typing.
type_cgraph(cg_leaf(Func, [])) :-
	type_func(Func).

type_cgraph(cgraph(Func, Calls)) :-
	type_cgraph_list(Calls),
	!,
	type_func(Func).

type_cgraph(cg_fix(FL, Calls)) :-
	type_cgraph_list(Calls),
	!,
	type_fixpoint(FL).

type_cgraph_list([]).
type_cgraph_list([F|FL]) :-
	type_cgraph(F),
	!,
	type_cgraph_list(FL).

%% Function typing, we have to type each rule.
type_func(Function) :-
	get_rules(Function, Rules),
	infer_type_rules(Rules, RulesType),
	get_general_type(RulesType, Type),
	type_check_save(Function, Type).

%% Fixpoint typing. We replace each function in the fixpoint with a
%% fresh var, and then we type the whole set (sequentially). This way
%% the types should be constrained to the right type (sic)
type_fixpoint(FL) :-
	build_fp_dicc(FL, Dicc),
	type_fixpoint_comp(FL, Dicc),
	type_check_save_fp(Dicc).

%% 
build_fp_dicc([], []).
build_fp_dicc([F|FR], [F-param(_)|R]) :-
	build_fp_dicc(FR, R).

type_fixpoint_comp([], _Dicc).
type_fixpoint_comp([F|FL], Dicc) :-
	get_rules(F, Rules),
	replace_fp_func(Rules, Dicc, NRules),
	infer_type_rules(NRules, RulesType),
	get_general_type(RulesType, _), %% Just unify
	type_fixpoint_comp(FL, Dicc).
	
%% 
replace_fp_func(Rules, Dicc, NRules) :-
	map_tree_list(Rules, free_fixpoint(Dicc), NRules).

free_fixpoint(X, Dicc, Res) :-
	\+ var(X),
	member(X-Res, Dicc),
	!.

%% Unbuild the dicc.
type_check_save_fp([]).
type_check_save_fp([F-param(Type)|R]) :-
	type_check_save(F, Type),
	!,
	type_check_save_fp(R).

%% Get the general unified type from a list of types.
get_general_type([Type], Type) :-
	!.
get_general_type([Type|Rest], Type) :-
	get_general_type(Rest, RType),
	!,
	unify_types(Type, RType).

%% Check an inferred type against the specified one, or in any other
%% case save it to the symbol table.
type_check_save(Function, Type) :-
	(
	    get_object_property(Function, type, DefType) ->
	    (
		instance(DefType, Type) ->
		true
	    ;
		throw(error(typer, type_def_mismatch(Function, Type, DefType)))
	    )
	;
	    set_object_property(Function, type, Type)
	).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

