%% Type of types. 
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
:- module(types, [check_type/1, type_arity/2, type_name/2,
                  map_type/3, map_types/1, pretty_type/2 ]).

:- use_package(hiord).

:- use_module(library(terms),      [atom_concat/2] ).

:- use_module(symtab, [get_object_property/3, replace_object_property/3]).

% TODO: Should export these ops to people who import this module.
:- include(type_ops).

%% Calc the type arity. Used extensibly on the translator.
type_arity(X, 0) :-
	var(X),
	!.

type_arity(_A --> B, N) :-
	!,
	type_arity(B, N1),
	N is N1 + 1.

type_arity(_, 0) :-
	!.

%% Return qid of a type.
type_name(Name<@_Args, Name).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% A map predicate for types. 

:- meta_predicate map_types(pred(2)).

map_types(Func) :-
	get_object_property(Name, type, Type),
	map_type(Type, Func, NType),
	replace_object_property(Name, type, NType),
	fail.
map_types(_F).

:- meta_predicate map_type(?, pred(2), ?).
map_type(T, Func, NT) :- 
	Func(T, T1),
	!,
	map_type(T1, Func, NT). % See map_tree

map_type(A, _F, A) :-
	var(A),
	!.

map_type(tid(M, N), _F, tid(M, N)) :-
	!.
map_type(tvar(A), _F, tvar(A)) :-
	!.

map_type(T1 --> T2, F, NT1 --> NT2) :-
	map_type(T1, F, NT1),
	map_type(T2, F, NT2),
	!.
map_type(T<@TL, F, NT<@NTL) :-
	map_type(T, F, NT),
	!,
	map_type_list(TL, F, NTL).

:- meta_predicate map_type_list(?, pred(2), ?).

map_type_list([], _F, []) :-
	!.
map_type_list([T|TL], Func, [NT|NTL]) :-
	map_type(T, Func, NT),
	!,
	map_type_list(TL, Func, NTL).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Check if Term is a valid type.
%% An atom is a type name.
check_type(tid(_, _)).

check_type(tv(_)).

check_type(X<@TVars) :-
	check_type(X),
	check_type_list(TVars).

check_type(A-->B) :-
	check_type(A),
	check_type(B).

check_type((_TClass=>T)) :-
	check_type(T).

check_type_list([]).
check_type_list([T|TL]) :-
	check_type(T),
	check_type_list(TL).

% :- pred pretty_type: type -> term
pretty_type(A, A) :-
	var(A),
	!.

pretty_type(T --> T1, R) :-
	pretty_type(T, TN),
	pretty_type(T1, T1N),
	R = (TN -> T1N).

%% Special builtins.
pretty_type(tid('Prelude', void)<@[], R) :-
	R = 'Prelude.()',
	!.
pretty_type(tid('Prelude', [])<@Args, R) :-
	!,
	pretty_type_list(Args, L),
	R = L.

pretty_type(tid('Prelude', 'tuple')<@[A,B], R) :-
	!,
	pretty_type_list([A,B], [Ar, Br]),
	R = (Ar, Br).

pretty_type(T<@Args, R) :-
	pretty_type(T, TN),
	pretty_type_list(Args, ArgsN),
	R =.. [TN|ArgsN].

pretty_type(tid(Mod, Name), B) :-
	atom_concat([Mod, '.', Name], B).

pretty_type_list([], []).
pretty_type_list([T|TL], [TT|TN]) :-
	pretty_type_list(TL, TN),
	pretty_type(T, TT).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:
